#include "ClusteringAlg.h"

#include <cmath>
#include <functional>
#include <iomanip>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>

using std::greater;
using std::cout;
using std::endl;
using std::setw;
using std::string;
using std::vector;

ClusteringAlg::ClusteringAlg(Event* event, Geometry* geo, Headers* runHeader):SystemAlg( event,geo,runHeader){
  m_event = event;
  m_geo = geo;
  m_runHeader = runHeader;

  // create the new System Alg Objects
  m_noise = new NoiseAlg(m_event, m_geo, m_runHeader);
  m_calibration = new CalibrationAlg(m_event, m_geo, m_runHeader);
  vector<SystemAlg*> algVector;

  algVector.push_back(m_noise);
  algVector.push_back(m_calibration);
  SystemAlg::setAlgs(algVector);

  m_firstEvent = true;

}

// initialize any variables here
bool ClusteringAlg::initialize(){
  SystemAlg::initialize(); // could remove this line
  return true;
}

// put any code you want executed per event here
bool ClusteringAlg::execute(){
  SystemAlg::execute(); // could remove this line

  if (m_firstEvent) {
    m_firstEvent = false;
    // loop over cells to verify that all cells have a positive noise value
    bool badNoise = false;
    for (int febno = m_runHeader->hec()->getFirstCh(); febno <= m_runHeader->hec()->getLastCh(); ++febno) {
      if (m_geo->hec_isConnected(febno)) {
        float noise = m_noise->hecNoise(febno);
        if (noise <= 0.) {
          cout << "*** warning: hec  cell " << febno << " has the noise value " << noise << " nA" << endl;
          badNoise = true;
        }
      }
    }
    for (int febno = m_runHeader->emec()->getFirstCh(); febno <= m_runHeader->emec()->getLastCh(); ++febno) {
      if (m_geo->emec_isConnected(febno)) {
        float noise = m_noise->emecNoise(febno);
        if (noise <= 0.) {
          cout << "*** warning: emec cell " << febno << " has the noise value " << noise << " nA" << endl;
          badNoise = true;
        }
      }
    }
    if (badNoise) cout << "*** warning: these cells will not be part of clusters!!!" << endl;
  }

  // febno keyed maps
  map<int, bool> aboveCellThresholdMap;
  map<int, bool> aboveNeighborThresholdMap;
  map<int, bool> aboveSeedThresholdMap;
  // signal over noise keyed seed cell febno map
  // greater<float> ensures that febnos are stored in decreasing order of the key
  // this will be used only for the original seed cell list
  multimap<float, int, greater<float> > seedCellMap;
  // seed cell vectors
  vector<int> seedCells;
  vector<int> nextSeedCells;

  // HEC
//   // debug
//   // choose an event number for debugging
//   int hecEventNumberDebug = 25;

  m_hecClusterMap.clear();
  if (m_hecEnable) {
    // clear clusters
    clearClusters();
    // clear maps and vectors
    aboveCellThresholdMap.clear();
    aboveNeighborThresholdMap.clear();
    aboveSeedThresholdMap.clear();
    seedCellMap.clear();
    seedCells.clear();
    nextSeedCells.clear();

    // loop over cells to initialize maps
    for (int febno = m_runHeader->hec()->getFirstCh(); febno <= m_runHeader->hec()->getLastCh(); ++febno) {
      if (m_geo->hec_isConnected(febno)) {
        // set default values for maps
        isNotInAnyCluster(febno);
        aboveCellThresholdMap[febno] = false;
        aboveNeighborThresholdMap[febno] = false;
        aboveSeedThresholdMap[febno] = false;

        // get signal
        float s = m_event->hecCell(febno)->signal();
        if (m_runHeader->hec()->getEUnit() == 1) s = m_calibration->hecADCTonA(febno, s);  // ADC to nA
        // get noise
        float noise = m_noise->hecNoise(febno);
        if (noise > 0.) {
          // categorize cells
          // the order at which this is done counts!
          if (s > m_hecSeedThreshold*noise) {
            // cells above seed threshold
            aboveSeedThresholdMap[febno] = true;
            seedCellMap.insert(multimap<float, int>::value_type(s/noise, febno));
            // create proto cluster
            int clusterID = createCluster();
            addToCluster(clusterID, febno);
          } else if (fabs(s) > m_hecNeighborThreshold*noise) {
            // cells above neighbor threshold
            aboveNeighborThresholdMap[febno] = true;
          } else if (fabs(s) > m_hecCellThreshold*noise) {
            // cells above cell threshold
            aboveCellThresholdMap[febno] = true;
          }
        }
      }
    }
    // set the initial seed cell list in order of decreasing signal/noise
    for (multimap<float, int>::iterator it = seedCellMap.begin(); it != seedCellMap.end(); it++) {
      nextSeedCells.push_back(it->second);
    }

//     // debug
//     if (m_event->hecEventNumber() == hecEventNumberDebug) {
//       cout << "ClusteringAlg::execute() hec: cells above seed threshold: ";
//       for (map<int, bool>::iterator it = aboveSeedThresholdMap.begin(); it != aboveSeedThresholdMap.end(); it++) {
//         if (it->second) cout << it->first << " ";
//       }
//       cout << endl;
//       cout << setw(5) << "febno" << setw(12) << "S/N" << endl;
//       for (multimap<float, int>::iterator it = seedCellMap.begin(); it != seedCellMap.end(); it++) {
//         cout << setw(5) << it->second << setw(12) << it->first << endl;
//       }
//       cout << "First ordered list of seed cells" << endl;
//       for (vector<int>::iterator it = nextSeedCells.begin(); it != nextSeedCells.end(); it++) {
//         cout << *it << " ";
//       }
//       cout << endl;
//       cout << "ClusteringAlg::execute() hec: cells above neighbor threshold: ";
//       for (map<int, bool>::iterator it = aboveNeighborThresholdMap.begin(); it != aboveNeighborThresholdMap.end(); it++) {
//         if (it->second) cout << it->first << " ";
//       }
//       cout << endl;
//       cout << "ClusteringAlg::execute() hec: cells above cell threshold: ";
//       for (map<int, bool>::iterator it = aboveCellThresholdMap.begin(); it != aboveCellThresholdMap.end(); it++) {
//         if (it->second) cout << it->first << " ";
//       }
//       cout << endl;
//       for (map<int, vector<int> >::iterator it = m_clusterMap.begin(); it != m_clusterMap.end(); it++) {
//         vector<int> cluster = it->second;
//         cout << "ClusteringAlg::execute() hec: cluster " << it->first << ": ";
//         for (vector<int>::iterator ic = cluster.begin(); ic != cluster.end(); ic++) {
//           cout << *ic << "  ";
//         }
//         cout << endl;
//       }
//     }

    // start the clustering iteration
    m_hecNumberOfIterations = 0;
    int itNMax = 100;
    while (!nextSeedCells.empty() && m_hecNumberOfIterations < itNMax) {
      m_hecNumberOfIterations++;

//       // debug
//       if (m_event->hecEventNumber() == hecEventNumberDebug) cout << "ClusteringAlg::execute() hec: start iteration " << m_hecNumberOfIterations << endl;

      // swap seedCells and nextSeedCells, and then clear nextSeedCells
      seedCells.swap(nextSeedCells);
      nextSeedCells.clear();
      
      if (m_hecSortingAtEachIteration) {
        // sort seed cells in order of decreasing signal/noise
        seedCellMap.clear();
        for (vector<int>::iterator it = seedCells.begin(); it != seedCells.end(); it++) {
          int febno = *it;
          // get signal
          float s = m_event->hecCell(febno)->signal();
          if (m_runHeader->hec()->getEUnit() == 1) s = m_calibration->hecADCTonA(febno, s);  // ADC to nA
          // get noise
          float noise = m_noise->hecNoise(febno);
          seedCellMap.insert(multimap<float, int>::value_type(s/noise, febno));
        }
        seedCells.clear();
        for (multimap<float, int>::iterator it = seedCellMap.begin(); it != seedCellMap.end(); it++) {
          seedCells.push_back(it->second);
        }
      }

      // loop over seedCells
      for (vector<int>::iterator it = seedCells.begin(); it != seedCells.end(); it++) {
        int febno = *it;
        int clusterID = m_clusterIDMap[febno];

//         // debug
//         if (m_event->hecEventNumber() == hecEventNumberDebug) {
//           // get signal
//           float s = m_event->hecCell(febno)->signal();
//           if (m_runHeader->hec()->getEUnit() == 1) s = m_calibration->hecADCTonA(febno, s);  // ADC to nA
//           // get noise
//           float noise = m_noise->hecNoise(febno);
//           cout << "ClusteringAlg::execute() hec: seed cell " << febno 
//                << ", layer " << m_geo->hec_iz(febno) 
//                << ", cluster " << clusterID
//                << ", S/N " << s/noise << endl;
//         }

        // get the seed cell neighbors
        vector<int> neighbors;
        if (m_hecClusteringIs2D) {        
          // 2D neighbors
          neighbors = m_geo->hec_neighbors(febno, m_geo->hec_iz(febno));
        } else {
          // 3D neighbors
          neighbors = m_geo->hec_neighbors(febno);
        }          

        // loop over the neighbors
        for (vector<int>::iterator itn = neighbors.begin(); itn != neighbors.end(); itn++) {
          int febnoNeighbor = *itn;

//           // debug
//           if (m_event->hecEventNumber() == hecEventNumberDebug) cout << "ClusteringAlg::execute() hec: neigbor " << febnoNeighbor << endl;
         
          // treat the various neighbor cases
          // here the order of the cases does not matter since the lists are exclusive
          // but the cell order within a given list does matter!
          if (aboveCellThresholdMap[febnoNeighbor]) {
            // case of neighbor above cell threshold
            // add to seed cluster if not in any clusters
            if (notInAnyCluster(febnoNeighbor)) addToCluster(clusterID, febnoNeighbor);
          } else if (aboveNeighborThresholdMap[febnoNeighbor]) {
            // case of neighbor above neighbor threshold
            if (notInAnyCluster(febnoNeighbor)) {
              // if not in any clusters add to seed cluster and to nextSeedCells
              addToCluster(clusterID, febnoNeighbor);
              nextSeedCells.push_back(febnoNeighbor);
            } else {
              int clusterID2 = m_clusterIDMap[febnoNeighbor];
              if (clusterID != clusterID2) {
                // if in a different cluster then merge clusters
                mergeClusters(clusterID, clusterID2);
              }
            }
          } else if (aboveSeedThresholdMap[febnoNeighbor]) {
            // case of neighbor above seed threshold, hence belonging to a cluster
            int clusterID2 = m_clusterIDMap[febnoNeighbor];
            if (clusterID != clusterID2) {
              // if in a different cluster then merge clusters
              mergeClusters(clusterID, clusterID2);
            }
          }

//           // debug
//           if (m_event->hecEventNumber() == hecEventNumberDebug) {
//             for (map<int, vector<int> >::iterator it = m_clusterMap.begin(); it != m_clusterMap.end(); it++) {
//               vector<int> cluster = it->second;
//               cout << "ClusteringAlg::execute() hec: cluster " << it->first << ": ";
//               for (vector<int>::iterator ic = cluster.begin(); ic != cluster.end(); ic++) {
//                 cout << *ic << "  ";
//               }
//               cout << endl;
//             }
//           }

        }
      }
    }
    if (!nextSeedCells.empty() && m_hecNumberOfIterations >= itNMax) 
      cout << "ClusteringAlg::execute() hec: stopped after " << itNMax << " iterations!!" << endl;
  }
  // store hec cluster maps
  m_hecClusterIDMap = m_clusterIDMap;
  m_hecClusterMap = m_clusterMap;

  // EMEC
//   // debug
//   // choose an event number for debugging
//   int emecEventNumberDebug = 25;

  m_emecClusterMap.clear();
  if (m_emecEnable) {
    // clear clusters
    clearClusters();
    // clear maps and vectors
    aboveCellThresholdMap.clear();
    aboveNeighborThresholdMap.clear();
    aboveSeedThresholdMap.clear();
    seedCellMap.clear();
    seedCells.clear();
    nextSeedCells.clear();

    // loop over cells to initialize maps
    for (int febno = m_runHeader->emec()->getFirstCh(); febno <= m_runHeader->emec()->getLastCh(); ++febno) {
      if (m_geo->emec_isConnected(febno)) {
        // set default values for maps
        isNotInAnyCluster(febno);
        aboveCellThresholdMap[febno] = false;
        aboveNeighborThresholdMap[febno] = false;
        aboveSeedThresholdMap[febno] = false;

        // get signal
        float s = m_event->emecCell(febno)->signal();
        if (m_runHeader->emec()->getEUnit() == 1) s = m_calibration->emecADCTonA(febno, s);  // ADC to nA
        // get noise
        float noise = m_noise->emecNoise(febno);
        if (noise > 0.) {
          // categorize cells
          // the order at which this is done counts!
          if (s > m_emecSeedThreshold*noise) {
            // cells above seed threshold
            aboveSeedThresholdMap[febno] = true;
            seedCellMap.insert(multimap<float, int>::value_type(s/noise, febno));
            // create proto cluster
            int clusterID = createCluster();
            addToCluster(clusterID, febno);
          } else if (fabs(s) > m_emecNeighborThreshold*noise) {
            // cells above neighbor threshold
            aboveNeighborThresholdMap[febno] = true;
          } else if (fabs(s) > m_emecCellThreshold*noise) {
            // cells above cell threshold
            aboveCellThresholdMap[febno] = true;
          }
        }
      }
    }
    // set the initial seed cell list in decreasing order of signal/noise
    for (multimap<float, int>::iterator it = seedCellMap.begin(); it != seedCellMap.end(); it++) {
      nextSeedCells.push_back(it->second);
    }

//     // debug
//     if (m_event->emecEventNumber() == emecEventNumberDebug) {
//       cout << "ClusteringAlg::execute() emec: cells above seed threshold: ";
//       for (map<int, bool>::iterator it = aboveSeedThresholdMap.begin(); it != aboveSeedThresholdMap.end(); it++) {
//         if (it->second) cout << it->first << " ";
//       }
//       cout << endl;
//       cout << setw(5) << "febno" << setw(12) << "S/N" << endl;
//       for (multimap<float, int>::iterator it = seedCellMap.begin(); it != seedCellMap.end(); it++) {
//         cout << setw(5) << it->second << setw(12) << it->first << endl;
//       }
//       cout << "First ordered list of seed cells" << endl;
//       for (vector<int>::iterator it = nextSeedCells.begin(); it != nextSeedCells.end(); it++) {
//         cout << *it << " ";
//       }
//       cout << endl;
//       cout << "ClusteringAlg::execute() emec: cells above neighbor threshold: ";
//       for (map<int, bool>::iterator it = aboveNeighborThresholdMap.begin(); it != aboveNeighborThresholdMap.end(); it++) {
//         if (it->second) cout << it->first << " ";
//       }
//       cout << endl;
//       cout << "ClusteringAlg::execute() emec: cells above cell threshold: ";
//       for (map<int, bool>::iterator it = aboveCellThresholdMap.begin(); it != aboveCellThresholdMap.end(); it++) {
//         if (it->second) cout << it->first << " ";
//       }
//       cout << endl;
//       for (map<int, vector<int> >::iterator it = m_clusterMap.begin(); it != m_clusterMap.end(); it++) {
//         vector<int> cluster = it->second;
//         cout << "ClusteringAlg::execute() emec: cluster " << it->first << ": ";
//         for (vector<int>::iterator ic = cluster.begin(); ic != cluster.end(); ic++) {
//           cout << *ic << "  ";
//         }
//         cout << endl;
//       }
//     }

    // start the clustering iteration
    m_emecNumberOfIterations = 0;
    int itNMax = 100;
    while (!nextSeedCells.empty() && m_emecNumberOfIterations < itNMax) {
      m_emecNumberOfIterations++;

//       // debug
//       if (m_event->emecEventNumber() == emecEventNumberDebug) cout << "ClusteringAlg::execute() emec: start iteration " << m_emecNumberOfIterations << endl;

      // swap seedCells and nextSeedCells, and then clear nextSeedCells
      seedCells.swap(nextSeedCells);
      nextSeedCells.clear();

      if (m_emecSortingAtEachIteration) {
        // sort seed cells in order of decreasing signal/noise
        seedCellMap.clear();
        for (vector<int>::iterator it = seedCells.begin(); it != seedCells.end(); it++) {
          int febno = *it;
          // get signal
          float s = m_event->emecCell(febno)->signal();
          if (m_runHeader->emec()->getEUnit() == 1) s = m_calibration->emecADCTonA(febno, s);  // ADC to nA
          // get noise
          float noise = m_noise->emecNoise(febno);
          seedCellMap.insert(multimap<float, int>::value_type(s/noise, febno));
        }
        seedCells.clear();
        for (multimap<float, int>::iterator it = seedCellMap.begin(); it != seedCellMap.end(); it++) {
          seedCells.push_back(it->second);
        }
      }
      
      // loop over seedCells
      for (vector<int>::iterator it = seedCells.begin(); it != seedCells.end(); it++) {
        int febno = *it;
        int clusterID = m_clusterIDMap[febno];

//         // debug
//         if (m_event->emecEventNumber() == emecEventNumberDebug) {
//           // get signal
//           float s = m_event->emecCell(febno)->signal();
//           if (m_runHeader->emec()->getEUnit() == 1) s = m_calibration->emecADCTonA(febno, s);  // ADC to nA
//           // get noise
//           float noise = m_noise->emecNoise(febno);
//           cout << "ClusteringAlg::execute() emec: seed cell " << febno 
//                << ", layer " << m_geo->emec_iz(febno) 
//                << ", cluster " << clusterID
//                << ", S/N " << s/noise << endl;
//         }

        // get the seed cell neighbors
        vector<int> neighbors;
        if (m_emecClusteringIs2D) {        
          // 2D neighbors
          neighbors = m_geo->emec_neighbors(febno, m_geo->emec_iz(febno));
        } else {
          // 3D neighbors
          neighbors = m_geo->emec_neighbors(febno);
        }          

        // loop over the neighbors
        for (vector<int>::iterator itn = neighbors.begin(); itn != neighbors.end(); itn++) {
          int febnoNeighbor = *itn;

//           // debug
//           if (m_event->emecEventNumber() == emecEventNumberDebug) cout << "ClusteringAlg::execute() emec: neigbor " << febnoNeighbor << endl;
         
          // treat the various neighbor cases
          // here the order of the cases does not matter since the lists are exclusive
          // but the cell order within a given list does matter!
          if (aboveCellThresholdMap[febnoNeighbor]) {
            // case of neighbor above cell threshold
            // add to seed cluster if not in any clusters
            if (notInAnyCluster(febnoNeighbor)) addToCluster(clusterID, febnoNeighbor);
          } else if (aboveNeighborThresholdMap[febnoNeighbor]) {
            // case of neighbor above neighbor threshold
            if (notInAnyCluster(febnoNeighbor)) {
              // if not in any clusters add to seed cluster and to nextSeedCells
              addToCluster(clusterID, febnoNeighbor);
              nextSeedCells.push_back(febnoNeighbor);
            } else {
              int clusterID2 = m_clusterIDMap[febnoNeighbor];
              if (clusterID != clusterID2) {
                // if in a different cluster then merge clusters
                mergeClusters(clusterID, clusterID2);
              }
            }
          } else if (aboveSeedThresholdMap[febnoNeighbor]) {
            // case of neighbor above seed threshold, hence belonging to a cluster
            int clusterID2 = m_clusterIDMap[febnoNeighbor];
            if (clusterID != clusterID2) {
              // if in a different cluster then merge clusters
              mergeClusters(clusterID, clusterID2);
            }
          }

//           // debug
//           if (m_event->emecEventNumber() == emecEventNumberDebug) {
//             for (map<int, vector<int> >::iterator it = m_clusterMap.begin(); it != m_clusterMap.end(); it++) {
//               vector<int> cluster = it->second;
//               cout << "ClusteringAlg::execute() emec: cluster " << it->first << ": ";
//               for (vector<int>::iterator ic = cluster.begin(); ic != cluster.end(); ic++) {
//                 cout << *ic << "  ";
//               }
//               cout << endl;
//             }
//           }

        }
      }
    }
    if (!nextSeedCells.empty() && m_emecNumberOfIterations >= itNMax) 
      cout << "ClusteringAlg::execute() emec: stopped after " << itNMax << " iterations!!" << endl;
  }
  // store emec cluster maps
  m_emecClusterIDMap = m_clusterIDMap;
  m_emecClusterMap = m_clusterMap;

  return true;
}

// put any code here that you want executed after all the events have been processed.
bool ClusteringAlg::finalize(){
  SystemAlg::finalize(); // could remove this line
  return true;
}

// enable the search of clusters for the hec
void ClusteringAlg::hecEnable(string noiseFileName, bool clusteringIs2D, bool sortingAtEachIteration, float cellThreshold, float neighborThreshold, float seedThreshold){
  // set the enable flag
  m_hecEnable = true;
  // set the clustering type flags
  m_hecClusteringIs2D = clusteringIs2D;
  m_hecSortingAtEachIteration = sortingAtEachIteration;
  // set the clustering thresholds
  m_hecCellThreshold = cellThreshold;
  m_hecNeighborThreshold = neighborThreshold;
  m_hecSeedThreshold = seedThreshold;
  cout << "hec clustering" << endl;
  if (m_hecClusteringIs2D) {
    cout << "  2d";
  } else {
    cout << "  3d";
  }
  cout << " with seed cells sorted in decreasing order of S/N";
  if (m_hecSortingAtEachIteration) {
    cout << " at each iteration" << endl;
  } else {
    cout << " only at the first iteration" << endl;
  }
  cout << "  seed threshold:     " << m_hecSeedThreshold     << " sigma(noise)" << endl;
  cout << "  neighbor threshold: " << m_hecNeighborThreshold << " sigma(noise)" << endl;
  cout << "  cell threshold:     " << m_hecCellThreshold     << " sigma(noise)" << endl;

  if (noiseFileName == "") {
    // get the default file name
    // hec_nnnnn_noise.dat
    std::ostringstream os;
    os << "hec_" << m_runHeader->global()->getRunNo() << "_noise.dat";
    noiseFileName = os.str();
  }
  // load the noise file
  cout << "ClusteringAlg:: loading noise file " << noiseFileName << endl;
  m_noise->hecReadNoiseFile(noiseFileName);
  // calibration file must be given if data is in ADC
  if (m_runHeader->hec()->getEUnit() == 1) { // ADC
    if (m_hecCalibrationFileName == "") {
      cout << "The ntuple data is in ADC: you must specify a calibration file with hecCalibrationFileName" << endl;
      exit(1);
    } else {  // load the calibration coefficients
      m_calibration->hecLoad(m_hecCalibrationFileName);
    }
  }
}

// enable the search of clusters for the emec
void ClusteringAlg::emecEnable(string noiseFileName, bool clusteringIs2D, bool sortingAtEachIteration, float cellThreshold, float neighborThreshold, float seedThreshold){
  // set the enable flag
  m_emecEnable = true;
  // set the clustering type flags
  m_emecClusteringIs2D = clusteringIs2D;
  m_emecSortingAtEachIteration = sortingAtEachIteration;
  // set the clustering thresholds
  m_emecCellThreshold = cellThreshold;
  m_emecNeighborThreshold = neighborThreshold;
  m_emecSeedThreshold = seedThreshold;
  cout << "emec clustering" << endl;
  if (m_emecClusteringIs2D) {
    cout << "  2d";
  } else {
    cout << "  3d";
  }
  cout << " with seed cells sorted in decreasing order of S/N";
  if (m_emecSortingAtEachIteration) {
    cout << " at each iteration" << endl;
  } else {
    cout << " only at the first iteration" << endl;
  }
  cout << "  seed threshold:     " << m_emecSeedThreshold     << " sigma(noise)" << endl;
  cout << "  neighbor threshold: " << m_emecNeighborThreshold << " sigma(noise)" << endl;
  cout << "  cell threshold:     " << m_emecCellThreshold     << " sigma(noise)" << endl;

  if (noiseFileName == "") {
    // get the default file name
    // emec_nnnnn_noise.dat
    std::ostringstream os;
    os << "emec_" << m_runHeader->global()->getRunNo() << "_noise.dat";
    noiseFileName = os.str();
  }
  // load the noise file
  cout << "ClusteringAlg:: loading noise file " << noiseFileName << endl;
  m_noise->emecReadNoiseFile(noiseFileName);
  // calibration file must be given if data is in ADC
  if (m_runHeader->emec()->getEUnit() == 1) { // ADC
    if (m_emecCalibrationFileName == "") {
      cout << "The ntuple data is in ADC: you must specify a calibration file with emecCalibrationFileName" << endl;
      exit(1);
    } else {  // load the calibration coefficients
      m_calibration->emecLoad(m_emecCalibrationFileName);
    }
  }
}

// clears the vector of clusters and the cluster ID map
void ClusteringAlg::clearClusters(){
  // reset next available cluster ID
  m_clusterID = 0;
  // clear cluster ID map
  m_clusterIDMap.clear();
  // clear cluster map
  m_clusterMap.clear();
}

// create a cluster, and return its ID
int ClusteringAlg::createCluster(){
  vector<int> clusterV;
  m_clusterMap[m_clusterID] = clusterV;
  return m_clusterID++;
}

// add a cell to a cluster
// prints a warning and does nothing if cluster does not exist
// does not check if the cell is already in the cluster...
void ClusteringAlg::addToCluster(int clusterID, int febno){
  map<int, vector<int> >::iterator it = m_clusterMap.find(clusterID);
  if (it != m_clusterMap.end()) {
    // add febno to cluster
    vector<int> cluster = it->second;
    cluster.push_back(febno);
    m_clusterMap[clusterID] = cluster;
    // update cluster ID map
    m_clusterIDMap[febno] = clusterID;
  } else {
    cout << "WARNING: inexistant cluster ID in addToCluster: "
         << clusterID
         << ".  No action taken" << endl;
  }
}
  
// merge clusters
// does nothing if both clusters have the same ID's
// prints a warning and does nothing if one cluster or both clusters do not exist
// cluster2 becomes part of cluster1 (cluster2 is then deleted)
void ClusteringAlg::mergeClusters(int clusterID1, int clusterID2){

  if (clusterID1 != clusterID2) {
    map<int, vector<int> >::iterator it1 = m_clusterMap.find(clusterID1);
    map<int, vector<int> >::iterator it2 = m_clusterMap.find(clusterID2);
    map<int, vector<int> >::iterator itEnd = m_clusterMap.end();
    if (it1 != itEnd && it2 != itEnd){
      vector<int> cluster1 = it1->second;
      vector<int> cluster2 = it2->second;
      for (vector<int>::iterator it = cluster2.begin(); it != cluster2.end(); it++) {
        int febno = *it;
        // add cells of cluster2 into cluster1
        cluster1.push_back(febno);
        // update cluster ID map
        m_clusterIDMap[febno] = clusterID1;
      }
      // update cluster1 in the cluster map
      m_clusterMap[clusterID1] = cluster1;
      // erase cluster2 from the cluster map
      m_clusterMap.erase(it2);
    } else {
      cout << " WARNING: inexistant cluster ID in mergeClusters: ";
      if (it1 == itEnd) cout << clusterID2;
      if (it1 == itEnd && it2 == itEnd) cout << " and ";
      if (it2 == itEnd) cout << clusterID2;
      cout << endl;
    }
  }
}


